//! API endpoints and request handlers.

use std::convert::Infallible;
use std::net::SocketAddr;
use std::sync::Arc;

use http_body_util::{BodyExt as _, Full};
use hyper::body::Bytes;
use hyper::server::conn::http1;
use hyper::service::service_fn;
use hyper_util::rt::TokioIo;
use hyper::Method;
use puppy::Context;
use serde_json::{json, Value};
use tokio::net::TcpListener;
use tracing::{error, info, info_span, trace_span, Instrument as _};

use crate::sig::{Signer, Verdict, Verifier, VERIFIER_MOUNT};

use self::error::Message;

// A simple macro for returning an error message.
macro_rules! fuck {
    ($code:literal: $($arg:tt)*) => {
        return Err(crate::api::error::Message {
            status: $code,
            error: format!($($arg)*),
            detail: None,
        })
    };
}

// Makes a response.
macro_rules! respond {
    ($($arg:tt)*) => {
        crate::api::Resp {
            $($arg)*,
            .. crate::api::Resp::default()
        }.into()
    };
}

/// Parameters for a response
struct Resp<'s> {
    body: Option<&'s Value>,
    kind: &'s str,
    code: u16,
}

impl<'s> Default for Resp<'s> {
    fn default() -> Self {
        Resp {
            body: None,
            kind: "application/json",
            code: 200,
        }
    }
}

impl<'a> From<Resp<'a>> for Response {
    fn from(Resp { body, kind, code }: Resp<'_>) -> Response {
        let resp = Response::<()>::builder()
            .status(code)
            .header("content-type", kind);
        resp.body(match body {
            Some(data) => Full::new(serde_json::to_vec(&data).unwrap().into()),
            None => Full::new(Bytes::default()),
        })
        .unwrap()
    }
}

pub mod ap;
pub mod wf;

type Request = hyper::Request<hyper::body::Incoming>;
type Response<T = Full<Bytes>> = hyper::Response<T>;

/// Initialize the http server loop.
pub async fn start(context: Context) -> Result<(), Box<dyn std::error::Error + Send + Sync>> {
    let addr = SocketAddr::from(([127, 0, 0, 1], context.config().port));

    let listener = TcpListener::bind(addr).await?;
    let verifier = Arc::new(Verifier::load(context.config()));

    loop {
        let (stream, _) = listener.accept().await?;
        let io = TokioIo::new(stream);

        let cx = context.clone();
        let verifier = verifier.clone();

        tokio::spawn(async move {
            let service = service_fn(|req| {
                let user_agent = req
                    .headers()
                    .get("user-agent")
                    .and_then(|h| h.to_str().ok());
                let span = info_span!(
                    "request",
                    target = format!("{} {}", req.method().as_str(), req.uri().to_string()),
                    "user-agent" = user_agent,
                );
                handle(req, &verifier, cx.clone()).instrument(span)
            });
            if let Err(err) = http1::Builder::new().serve_connection(io, service).await {
                error!("Error serving connection: {:?}", err);
            }
        });
    }
}

// A parsed HTTP request for easy handling.
struct Req<'a> {
    method: &'a Method,
    body: Bytes,
    // The content-types in the accept header
    accept: Vec<&'a str>,
    // URI bits
    params: Vec<(&'a str, &'a str)>,
    path: Vec<&'a str>,
}

impl Req<'_> {
    /// Get the path segments (non-empty parts of the path string separated by the '/' character).
    fn path(&self) -> &[&str] {
        &self.path
    }
    /// Turn an HTTP request into a more simple form so we can process it more easily.
    fn simplify<'x>(r: &'x http::Request<Bytes>) -> Req<'x> {
        let path: Vec<&str> = r
            .uri()
            .path()
            .split('/')
            .filter(|s| !s.is_empty())
            .collect();
        let params: Vec<(&str, &str)> = r
            .uri()
            .query()
            .into_iter()
            .flat_map(|s| s.split('&'))
            .filter_map(|s| s.split_once('='))
            .collect();
        let accept = r
            .headers()
            .iter()
            .find_map(|(k, v)| (k == "accept").then_some(v))
            .and_then(|val| val.to_str().ok())
            .iter()
            .flat_map(|s| s.split(' '))
            .filter(|s| !s.is_empty())
            .collect();
        Req {
            method: r.method(),
            body: r.body().clone(),
            accept,
            params,
            path,
        }
    }
}

/// The request handler.
async fn handle(req: Request, verifier: &Verifier, cx: Context) -> Result<Response, Infallible> {
    // We need to fetch the entire body of the request for signature validation, because that involves making
    // a digest of the request body in some cases.
    // TODO: defer loading the body until it is needed.
    let request = {
        let (req, body) = req.into_parts();
        let Ok(body) = body.collect().await.map(|b| b.to_bytes()) else {
            panic!("could not collect body!");
        };
        http::Request::from_parts(req, body)
    };

    // Simplified representation of a request, so we can pattern match on it more easily in the dispatchers.
    let req = Req::simplify(&request);

    // We'll use the path to pick where specifically to send the request.
    // Check request signature at the door. Even if it isn't needed for a particular endpoint, failing fast
    // with a clear error message will save anyone trying to get *their* signatures implementation a major
    // headache.
    let res = match verifier.verify(&request).await {
        // If the request was signed and the signature was accepted, they can access the protected endpoints.
        Verdict::Verified(sig) => dispatch_signed(cx, &verifier, &req, sig).await,
        // Unsigned requests can see a smaller subset of endpoints, most notably the verification actor.
        Verdict::Unsigned => dispatch_public(cx, &verifier, &req).await,
        // If a signature was provided *but it turned out to be unverifiable*, show them the error message.
        Verdict::Rejected { reason, signature_str } => Err(Message {
            error: String::from("signature verification failed for request"),
            status: 403,
            detail: Some(json!({
                "signature": signature_str,
                "reason": reason,
            })),
        }),
    };

    // If one of the endpoints gave us an error message, we convert that into a response and then
    // serve it to the client. In either case, we just serve a response.
    let response = res.unwrap_or_else(|msg| {
        info!("{}: {msg}", msg.status);
        req.error(msg)
    });

    Ok(response)
}

const POST: &Method = &Method::POST;
const GET: &Method = &Method::GET;

/// Handle a signed and verified request.
///
/// This function is where all requests to a protected endpoint have to go through. If the request
/// was signed but does not target a protected endpoint, this function will fall back to the
/// [`dispatch_public`] handler.
#[tracing::instrument(level = "DEBUG", target = "router", skip_all)]
async fn dispatch_signed(
    cx: Context,
    verifier: &Verifier,
    req: &Req<'_>,
    sig: Signer,
) -> Result<Response, Message> {
    match (req.method, req.path()) {
        // Viewing ActivityPub objects requires a signed request, i.e. "authorized fetch".
        // The one exception for this is `/s/request-verifier`, which is where the request
        // verification actor lives.
        (GET, ["o", ulid]) => ap::serve_object(&cx, ulid),
        // POSTs to an actor's inbox need to be signed to prevent impersonation.
        (POST, ["o", ulid, "inbox"]) => ap::inbox(&cx, ulid, sig, &req.body).await,
        // Try the resources for which no signature is required as well.
        _ => dispatch_public(cx, verifier, req).await,
    }
}

/// Dispatch `req` to an unprotected endpoint. If the requested path does not exist, the
/// function will return a 404 response. If the path *does* exist, but the signature is not
/// valid, they will also get a 404.
#[tracing::instrument(level = "DEBUG", target = "router", skip_all)]
async fn dispatch_public(
    cx: Context,
    verifier: &Verifier,
    req: &Req<'_>,
) -> Result<Response, Message> {
    match (req.method, req.path()) {
        (GET, ["proxy"]) => ap::proxy(&cx, &req.params).await,
        (GET, ["outbox"]) => ap::outbox(&cx, &req.params).await,
        (GET, [".well-known", "webfinger"]) => wf::resolve(&cx, &req.params),
        // TODO: nicer solution for this
        (GET, VERIFIER_MOUNT) => Ok(ap::serve_verifier_actor(&verifier)),
        _ => fuck!(404: "not found"),
    }
}

mod error {
    //! Pre-baked error responses.

    use serde_json::{json, Value};
    use super::Response;

    /// An error message shown to an end user of the API.
    #[derive(Debug)]
    pub struct Message {
        /// The main error message.
        pub error: String,
        /// Only shown if the `accept` header included json.
        pub detail: Option<Value>,
        /// The status code for the response.
        pub status: u16,
    }

    impl std::fmt::Display for Message {
        fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
            self.error.fmt(f)
        }
    }

    impl super::Req<'_> {
        /// Generate an error response for the request.
        pub fn error(&self, err: Message) -> Response {
            let resp = Response::<()>::builder().status(err.status);
            // If the accept header wants json, we will give them a nice structured error
            // message. Otherwise, we throw a short bit of text at them.
            if self.accepts_json() {
                let json = json!({
                    "error": err.error,
                    "details": err.detail,
                });
                let body = serde_json::to_vec_pretty(&json).unwrap();
                resp.header("content-type", "application/json")
                    .body(body.try_into().unwrap())
                    .unwrap()
            } else {
                resp.header("content-type", "text/plain")
                    .body(err.error.try_into().unwrap())
                    .unwrap()
            }
        }
        /// Check whether the requester wants json from us.
        pub fn accepts_json(&self) -> bool {
            fn is_json((k, v): (&str, &str)) -> bool {
                k == "application" && v.split('+').any(|p| p == "json")
            }
            self.accept
                .iter()
                .filter_map(|s| s.split_once('/'))
                .any(is_json)
        }
    }
}
