use chrono::Utc;
use http::Method;
use http_body_util::BodyExt as _;
use reqwest::Body;
use serde_json::{error, Value};
use derive_more::Display;
use tracing::{debug, error, info, instrument, warn};

use crate::{
    object::Activity,
    signatures::{SigningKey, Options},
    FetchError,
};

/// The name of the server software, used for generating the user agent string.
///
/// See also [`VERSION`].
pub const SOFTWARE: &str = "ActivityPuppy";
/// The current version of the server software, which is incorporated into the user agent string
/// for all outbound requests made by ActivityPuppy.
pub const VERSION: &str = "0.0.1-dev";

/// Content-type/accept header for ActivityPub requests.
pub const ACTIVITYPUB_TYPE: &str = "application/activity+json";

/// A client for sending ActivityPub and WebFinger requests with.
#[derive(Clone)]
pub struct Client {
    inner: reqwest::Client,
}

impl Client {
    /// Constructs a new federation client.
    pub fn new() -> Client {
        Client { inner: reqwest::Client::new() }
    }
    /// Deliver an [`Activity`] to a particular `inbox`.
    ///
    /// Note that in order for the request to be considered valid by most implementations, `key.owner`
    /// must equal `payload.actor`.
    #[instrument(target = "fetch.delivery", skip_all, fields(activity = payload.id, url = inbox, key = key.id))]
    pub async fn deliver(&self, key: &SigningKey, payload: &Activity, inbox: &str) {
        let system = Subsystem::Delivery;

        let body = serde_json::to_string(&payload.to_json_ld()).unwrap();
        let mut req = system
            .new_request(inbox)
            .unwrap()
            .method(Method::POST)
            .header("content-type", ACTIVITYPUB_TYPE)
            .body(body)
            .unwrap();

        key.sign_with_digest(Options::LEGACY, &mut req)
            .map_err(FetchError::Sig)
            .expect("signature generation to work")
            .commit(&mut req);

        let request = req.map(Body::from).try_into().unwrap();
        self.inner.execute(request).await.unwrap();
    }
    /// A high-level function to resolve a single ActivityPub ID using a signed request.
    #[instrument(target = "fetch.resolver", skip_all, fields(url = url, key = key.id))]
    pub async fn resolve(&self, key: &SigningKey, url: &str) -> Result<Value, FetchError> {
        let system = Subsystem::Resolver;

        let mut req = system
            .new_request(url)?
            .header("accept", ACTIVITYPUB_TYPE)
            .body(())
            .unwrap();

        match key.sign(Options::LEGACY, &req) {
            Ok(signature) => signature.commit(&mut req),
            Err(error) => {
                // This shouldn't happen during normal operation
                warn!("failed to sign request: {error}");
                return Err(FetchError::Sig(error));
            }
        }

        let request = req.map(|()| Body::default()).try_into()?;
        let response = self.inner.execute(request).await?;

        if response.status().is_success() {
            debug!("resolution successful");
            response.json().await.map_err(From::from)
        } else {
            let status = response.status().as_u16();
            let body = response.text().await?;
            debug!(status, "resolution failed: {body}");
            Err(FetchError::NotSuccess {
                url: url.to_string(),
                status,
                body,
            })
        }
    }
    /// Forwards a request and returns the raw response, so that it can be analyzed for debugging.
    ///
    /// It exists solely as a debugging tool!
    #[instrument(target = "fetch.devproxy", skip_all, fields(url, key = key.id))]
    pub async fn proxy(
        &self,
        key: &SigningKey,
        url: &str,
    ) -> Result<http::Response<String>, FetchError> {
        let system = Subsystem::DevProxy;

        let mut req = system
            .new_request(url)?
            .header("accept", ACTIVITYPUB_TYPE)
            .body(())
            .unwrap();

        key.sign(Options::LEGACY, &req)
            .expect("signing error")
            .commit(&mut req);

        let resp = self
            .inner
            .execute(req.map(|_| Body::default()).try_into().unwrap())
            .await?;

        let http_resp: http::Response<reqwest::Body> = resp.into();
        let (res, body) = http_resp.into_parts();
        let body = body.collect().await.unwrap().to_bytes();
        let http_resp =
            http::Response::from_parts(res, String::from_utf8_lossy(body.as_ref()).into_owned());
        Ok(http_resp)
    }
}

/// Identifies a specific subsystem that makes an outgoing request.
///
/// This allows us to precisely track each outgoing request, as well as generate a meaningful
/// user-agent header. It is also used to generate a "base request".
#[derive(Clone, Copy, Display)]
enum Subsystem {
    /// The subsystem that dereferences ActivityPub URLs to JSON values.
    ///
    /// In addition, the resolver is used for resolving webfinger handles to ActivityPub actors.
    #[display(fmt = "resolver")]
    Resolver,
    /// The subsystem responsible for delivering activities to inboxes.
    #[display(fmt = "delivery")]
    Delivery,
    /// For testing the resolver and signatures.
    #[display(fmt = "devproxy")]
    DevProxy,
}

impl Subsystem {
    /// Get the user agent string for the subsystem.
    fn user_agent(&self) -> String {
        format!("{SOFTWARE}/{VERSION} [{}]", self.as_str())
    }
    /// Get a str representation of this subsystem.
    #[inline]
    const fn as_str(self) -> &'static str {
        match self {
            Subsystem::Resolver => "resolver",
            Subsystem::Delivery => "delivery",
            Subsystem::DevProxy => "devproxy",
        }
    }
    /// Construct a new request for this subsystem.
    ///
    /// This will set the following headers, which are common to all requests made by the fetch
    /// system:
    ///
    /// - `user-agent`, which depends on the particular subsystem in use
    /// - `date`, which is generated from the current time
    /// - `host`, which is derived from `target`
    ///
    /// This function returns an error if the `target` is not a valid URI. It panics if the URI
    /// does not have a host specified.
    fn new_request(self, target: &str) -> Result<http::request::Builder, FetchError> {
        // Format our time like "Sun, 06 Nov 1994 08:49:37 GMT"
        const RFC_822: &str = "%a, %d %b %Y %H:%M:%S GMT";
        let date = Utc::now().format(RFC_822).to_string();

        let uri = target
            .parse::<http::Uri>()
            .map_err(|e| FetchError::InvalidURI {
                url: target.to_string(),
                error: e.to_string(),
            })?;

        let Some(host) = uri.host() else {
            // SECURITY: Refuse to resolve URLs to local resources using local keys.
            error!(target: "security", "refusing to resolve a relative URL: {target}");
            return Err(FetchError::InvalidURI {
                url: target.to_string(),
                error: "Relative URI".to_string(),
            });
        };

        let req = http::Request::builder()
            .uri(target)
            .header("user-agent", self.user_agent())
            .header("date", date)
            .header("host", host);

        Ok(req)
    }
}
