#![feature(iter_intersperse, yeet_expr, iterator_try_collect, try_blocks)]
use std::error::Error;

use derive_more::Display;
use serde_json::Value;

use object::Activity;
use signatures::SigningKey;

pub use http;

pub mod signatures;
pub mod object;

pub use client::Client;
mod client;

/// Deliver an activity to an inbox.
pub async fn deliver(key: &SigningKey, activity: &Activity, inbox: &str) {
    Client::new().deliver(key, &activity, inbox).await
}

/// Resolve an ActivityPub ID to a JSON value.
///
/// Note: This creates a new [`Client`] every time you call it, so if you're gonna call it more than just
/// a couple of times, create a `Client` and call its inherent methods instead.
pub async fn resolve(key: &SigningKey, target: &str) -> Result<Value, FetchError> {
    Client::new().resolve(key, target).await
}

/// Proxy a GET request through this server.
///
/// Should only be used for manually testing stuff.
pub async fn forward(key: &SigningKey, target: &str) -> Result<http::Response<String>, FetchError> {
    Client::new().proxy(key, target).await
}

/// Errors that may occur during the execution of HTTP request routines.
#[derive(Debug, Display)]
pub enum FetchError {
    /// Some error internal to the request sending process occurred.
    #[display(fmt = "internal error: {error} (url={url:?})")]
    Internal { url: Option<String>, error: String },
    /// The URI was not valid and therefore the request could not be made.
    #[display(fmt = "invalid uri: {error} (url={url})")]
    InvalidURI { url: String, error: String },
    /// A non-success status code was encountered.
    #[display(fmt = "non-2xx status code: {status} (url={url})")]
    NotSuccess {
        status: u16,
        url: String,
        body: String,
    },
    /// The JSON body of a response could not be loaded. The string inside is the error
    /// message produced by the JSON deserializer.
    #[display(fmt = "deserialization error: {}", self.0)]
    BadJson(String),
    /// A JSON-LD document could not be deserialized because it does not conform to our expectations.
    #[display(fmt = "parsing error: {}", self.0)]
    BadObject(String),
    /// An error that occurred while generating a signature for a a request.
    #[display(fmt = "signing error: {}", self.0)]
    Sig(String),
}

impl FetchError {
    /// Check whether the error is due to a 403 UNAUTHORIZED response status code.
    pub fn is_unauthorized(&self) -> bool {
        matches!(self, FetchError::NotSuccess { status: 403, .. })
    }
    /// Check whether the error is due to a 404 NOT FOUND response status code.
    pub fn is_not_found(&self) -> bool {
        matches!(self, FetchError::NotSuccess { status: 404, .. })
    }
}

#[doc(hidden)]
impl From<reqwest::Error> for FetchError {
    fn from(error: reqwest::Error) -> FetchError {
        match error.source().and_then(|e| e.downcast_ref()) {
            Some(e @ serde_json::Error { .. }) => FetchError::BadJson(e.to_string()),
            None => {
                let url = error.url().map(|u| u.to_string());
                FetchError::Internal { url, error: error.to_string() }
            }
        }
    }
}
