//! ActivityPub vocabulary as interpreted by ActivityPuppy.

use serde_json::{json, Value};
use derive_more::From;

pub use crate::signatures::Key as PublicKey;

#[derive(Debug)]
pub struct Activity<T = String> {
    pub id: String,
    pub actor: String,
    pub object: Box<Object>,
    pub kind: T,
}

impl<K> Activity<K> {
    pub fn to_json_ld(&self) -> Value
    where
        K: ToString,
    {
        json!({
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                { "Bite": "https://ns.mia.jetzt/as#Bite" },
            ],
            "id": self.id.to_string(),
            "actor": self.actor.to_string(),
            "object": self.object.to_json_ld(),
            "type": self.kind.to_string(),
        })
    }
}
impl Activity {
    pub fn from_json(mut json: Value) -> Result<Activity, String> {
        let Some(map) = json.as_object() else {
            do yeet "expected an object"
        };
        let Some(id) = map.get("id").and_then(|s| s.as_str()).map(str::to_owned) else {
            do yeet "missing `id` property"
        };
        let Some(actor) = map.get("actor").and_then(|s| s.as_str()).map(str::to_owned) else {
            do yeet format!("missing `actor` property for activity {id}")
        };
        let Some(kind) = map.get("type").and_then(|s| s.as_str()).map(str::to_owned) else {
            do yeet format!("missing `type` property for activity {id}")
        };
        // TODO: make this behave gracefully when we only get an ID.
        let Some(object) = json
            .get_mut("object")
            .map(Value::take)
            .map(Object::from_json)
            .transpose()?
            .map(Box::new)
        else {
            do yeet format!("missing or invalid `object` property for activity {id}")
        };
        Ok(Activity { id, actor, object, kind })
    }
}

/// An actor is an entity capable of producing Takes.
#[derive(Debug)]
pub struct Actor {
    /// The URL pointing to this object.
    pub id: String,
    /// Where others should send activities.
    pub inbox: String,
    /// Note: this maps to the `preferredUsername` property.
    pub account_name: String,
    /// Note: this maps to the `name` property.
    pub display_name: Option<String>,
    /// Public counterpart to the signing key used to sign activities
    /// generated by the actor.
    pub public_key: PublicKey,
}

impl Actor {
    pub fn to_json_ld(&self) -> Value {
        json!({
            "@context": [
                "https://www.w3.org/ns/activitystreams",
                "https://w3id.org/security/v1",
            ],
            "id": self.id.to_string(),
            "inbox": self.inbox.to_string(),
            "outbox": self.inbox.to_string().replace("inbox", "outbox"),
            "preferredUsername": self.account_name,
            "name": self.display_name,
            "type": "Person",
            "publicKey": {
                "id": self.public_key.id,
                "publicKeyPem": self.public_key.inner,
                "owner": self.id.to_string(),
            }
        })
    }
    pub fn from_json(json: Value) -> Result<Actor, String> {
        let Value::Object(map) = json else {
            do yeet format!("expected json object")
        };
        Ok(Actor {
            id: map
                .get("id")
                .ok_or("id is required")?
                .as_str()
                .ok_or("id must be a str")?
                .to_string(),
            inbox: map
                .get("inbox")
                .ok_or("inbox is required")?
                .as_str()
                .ok_or("inbox must be a str")?
                .to_string(),
            account_name: map
                .get("preferredUsername")
                .ok_or("preferredUsername is required")?
                .as_str()
                .ok_or("preferredUsername must be a str")?
                .to_string(),
            display_name: map.get("name").and_then(|v| v.as_str()).map(str::to_owned),
            public_key: map
                .get("publicKey")
                .cloned()
                .and_then(PublicKey::from_json)
                .ok_or("publicKey property could not be parsed")?,
        })
    }
}

#[derive(From, Debug)]
pub enum Object {
    #[from(ignore)]
    Id {
        id: String,
    },
    Activity(Activity),
    Actor(Actor),
    Note(Note),
}

impl Object {
    pub fn id(&self) -> &str {
        match self {
            Object::Activity(a) => &a.id,
            Object::Actor(a) => &a.id,
            Object::Note(n) => &n.id,
            Object::Id { id } => id,
        }
    }
    pub fn from_json(json: Value) -> Result<Object, String> {
        if let Value::String(id) = json {
            Ok(Object::Id { id })
        } else if let Value::Object(ref map) = json {
            match map.get("type").and_then(Value::as_str) {
                Some("System" | "Application" | "Person" | "Service") => {
                    Actor::from_json(json).map(Object::Actor)
                }
                Some("Create" | "Follow" | "Accept" | "Reject" | "Bite") => {
                    Activity::from_json(json).map(Object::Activity)
                }
                Some(kind) => Ok(Object::Note(Note {
                    id: map
                        .get("id")
                        .ok_or("id is required")?
                        .as_str()
                        .ok_or("id must be a str")?
                        .to_string(),
                    kind: kind.to_string(),
                    author: map
                        .get("attributedTo")
                        .ok_or("attributedTo is required")?
                        .as_str()
                        .ok_or("attributedTo must be a str")?
                        .to_string(),
                    content: map
                        .get("content")
                        .and_then(|v| v.as_str())
                        .map(str::to_owned),
                    summary: map
                        .get("summary")
                        .and_then(|v| v.as_str())
                        .map(str::to_owned),
                })),
                None => do yeet "could not determine type of object",
            }
        } else {
            Err(format!("expected a json object or an id, got {json:#?}"))
        }
    }
    pub fn to_json_ld(&self) -> Value {
        match self {
            Object::Id { id } => json!(id),
            Object::Activity(a) => a.to_json_ld(),
            Object::Actor(a) => a.to_json_ld(),
            Object::Note(Note {
                id,
                kind,
                content,
                summary,
                author,
            }) => json!({
                "to": [
                    "https://www.w3.org/ns/activitystreams#Public",
                ],
                "id": id.to_string(),
                "type": kind,
                "attributedTo": author,
                "content": content,
                "summary": summary,
            }),
        }
    }
}

#[derive(Debug)]
pub struct Note {
    pub id: String,
    pub author: String,
    pub content: Option<String>,
    pub summary: Option<String>,
    pub kind: String,
}
