//! Containment zone for the funny math that doesn't make much sense to puppy.
//!
//! This module provides ActivityPuppy's HTTP signatures implementation. The state of HTTP signatures implementations
//! is, to put it mildly, *een fucking kutzooi*. For historical reasons, no one implements it *exactly* right (much
//! like URI parsers). This implementation aims to be as broadly compatible as possible.
//!
//! The only non-deprecated [`Algorithm`] is [`"hs2019"`][HS2019], but not everyone implements it, because the initial
//! round of implementations of the spec were based on a draft, and [`"rsa-sha256"`][RSA_SHA256] is kinda the de facto
//! standard.
//!
//! # Behavior
//!
//! By default, puppy will sign with `algorithm="hs2019"` (using `(created)` and `(expires)` pseudo-headers), and retry
//! in legacy mode (using `algorithm="rsa-sha256"` with `date` header) if the signature gets rejected.
//!
//! Currently, `"hs2019"` is treated as equivalent to `"rsa-sha256"` for verification purposes. Support for elliptic
//! curve keys is planned, but not a priority.
//!
//! # Links
//!
//! More information about http signatures:
//!
//! - <https://swicg.github.io/activitypub-http-signature>
//! - <https://datatracker.ietf.org/doc/html/draft-cavage-http-signatures>

use std::path::Path;

use chrono::{DateTime, TimeDelta, Utc};
use http::{HeaderValue, Request};
use rsa::{
    pkcs8::{
        DecodePrivateKey, DecodePublicKey, EncodePrivateKey as _, EncodePublicKey as _, LineEnding,
    },
    sha2::Sha256,
    signature::{SignatureEncoding as _, Signer as _, Verifier as _},
    RsaPrivateKey,
};

use serde_json::{Map, Value};

use self::new::{decode, encode, sha256, IR};

/// Size of the RSA private keys puppy generates.
const KEY_SIZE: usize = 2048;

/// A key that can be used to verify a request signature.
pub type VerificationKey = Key<Public>;

/// A key that can be used to sign a request.
pub type SigningKey = Key<Private>;

/// A key used for authorized fetch.
///
/// It comes in several flavors:
///
/// - `Key` (`K` = [`String`]): PEM-encoded, can be turned into a JSON object.
/// - [`VerificationKey`] (`K` = [`Public`]): used as an input in the request signature validation process.
/// - [`SigningKey`] (`K` = [`Private`]): used as an input in the generation of a signed request.
#[derive(Debug)]
pub struct Key<K = String> {
    /// The `"id"` property of the public key, which should equal the `keyId` part of a signature.
    pub id: String,
    /// The `"owner"` property.
    pub owner: String,
    /// Maps to the `"publicKeyPem"` property of an actor's `"publicKey"` when (de)serializing, and when the
    /// key is used for doing signatures.
    pub inner: K,
}

impl Key {
    /// Tries to find the PEM-encoded public key from the result of fetching a key id.
    pub fn from_json(json: Value) -> Option<Key> {
        // First, we try the object itself.
        json.as_object().and_then(Key::from_map).or_else(|| {
            // Because of how mastodon deals with pubkey resolution, most implementations will serve the whole actor
            // object instead of just the key, so we try that first, because it is the de facto standard.
            json.get("publicKey")?.as_object().and_then(Key::from_map)
        })
    }
    /// Try to interpret the given map as a public key.
    fn from_map(map: &Map<String, Value>) -> Option<Key> {
        Some(Key {
            id: map.get("id")?.as_str().map(str::to_owned)?,
            owner: map.get("owner")?.as_str().map(str::to_owned)?,
            inner: map.get("publicKeyPem")?.as_str().map(str::to_owned)?,
        })
    }
    /// "Upgrade" a pem-encoded public key to a key that can actually be used for requests.
    ///
    /// The inverse of this is [`Key::serialize`], which turns `inner` back into a string.
    ///
    /// [`Key::serialize`]: Key::<Public>::serialize
    pub fn upgrade(self) -> Key<Public> {
        let inner = Public::decode_pem(&self.inner);
        Key {
            id: self.id,
            owner: self.owner,
            inner,
        }
    }
}

/// A key that can be used to generate signatures.
#[derive(Clone)]
pub struct Private(rsa::RsaPrivateKey);

impl Private {
    /// Generate a new keypair.
    pub fn gen() -> (Private, Public) {
        let mut rng = rand::thread_rng();
        let private = RsaPrivateKey::new(&mut rng, KEY_SIZE).unwrap();
        let public = private.to_public_key();
        (Private(private), Public(public))
    }
    /// Get the public counterpart to this key.
    pub fn get_public(&self) -> Public {
        Public(self.0.to_public_key())
    }
    /// Load a private key from a file on disk.
    pub fn load(path: impl AsRef<Path>) -> Option<Private> {
        use rsa::pkcs8::DecodePrivateKey;
        let path = path.as_ref();
        DecodePrivateKey::read_pkcs8_pem_file(path)
            .map(Private)
            .ok()
    }
    /// Store the private key at `path`.
    pub fn save(&self, path: impl AsRef<Path>) {
        use rsa::pkcs8::EncodePrivateKey;
        self.0
            .write_pkcs8_pem_file(path, LineEnding::default())
            .expect("writing a private key to a file should not fail")
    }
    /// PEM-encode the key PKCS#8 style.
    pub fn encode_pem(&self) -> String {
        self.0
            .to_pkcs8_pem(LineEnding::default())
            .unwrap()
            .to_string()
    }
    /// Decode the key from a PKCS#8 PEM-encoded string.
    pub fn decode_pem(pkcs8_pem: &str) -> Private {
        DecodePrivateKey::from_pkcs8_pem(&pkcs8_pem)
            .map(Private)
            .unwrap()
    }
}

/// A key that can be used to verify signatures.
#[derive(Clone)]
pub struct Public(rsa::RsaPublicKey);

impl Public {
    /// PEM-encode the public key in accordance with PKCS#8.
    pub fn encode_pem(&self) -> String {
        self.0
            .to_public_key_pem(LineEnding::default())
            .unwrap()
            .to_string()
    }
    /// Decode a PKCS#8 PEM-encoded public key from a string.
    pub fn decode_pem(pkcs8_pem: &str) -> Public {
        let doc = pem::parse(pkcs8_pem).unwrap();
        <rsa::RsaPublicKey as DecodePublicKey>::from_public_key_der(doc.contents())
            .map(Public)
            .unwrap()
    }
}

impl std::fmt::Debug for Public {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.encode_pem().fmt(f)
    }
}

impl SigningKey {
    /// Create a signature for `req` using the given options.
    pub fn sign<T>(&self, opt: Options, req: &Request<T>) -> Result<Signature, String> {
        IR::partial(&req, opt, |ir| ir.signed(self)?.to_signature())
    }
    /// Create a signature for `req` using the given algorithm, and calculate and attach the `digest` header to
    /// the request (if it doesn't already have one).
    ///
    /// This is required by most implementations when POSTing to an inbox.
    pub fn sign_with_digest<T>(
        &self,
        opt: Options,
        req: &mut Request<T>,
    ) -> Result<Signature, String>
    where
        T: AsRef<[u8]>,
    {
        // Calculate and insert digest if it isn't there yet, otherwise do nothing.
        let digest = format!("sha-256={}", encode(sha256(req.body())));
        req.headers_mut()
            .entry("digest")
            .or_insert_with(|| digest.try_into().unwrap());
        self.sign(opt, req)
    }
}

impl VerificationKey {
    /// Test the signature against three requirements:
    ///
    /// 1. The signature must not be expired.
    /// 2. The signature's `keyId` must be the same as `self`'s.
    /// 3. The `signed_str` must have been signed by the private counterpart of `self`.
    pub fn verify(&self, sig: &Signature) -> Result<(), String> {
        if sig.is_expired_at(Utc::now()) {
            do yeet format!("Signature is expired: (deadline was {})", sig.expires());
        }
        if sig.key_id() != &self.id {
            do yeet format! {
                "Mismatched key id; signature's key id is '{}', while presented key has '{}'",
                sig.key_id(),
                self.id
            };
        }
        if !sig.was_signed_by(&self.inner)? {
            do yeet "Signature was not generated by the presented key's private counterpart";
        }
        Ok(())
    }
    /// Encode a verification key so that it can be presented in a json.
    pub fn serialize(self) -> Key {
        let public_key_pem = self.inner.encode_pem();
        Key {
            id: self.id,
            owner: self.owner,
            inner: public_key_pem,
        }
    }
}

/// The algorithm to sign with.
///
/// Your two options are:
///
/// - [`hs2019`][HS2019], the *correct* option
/// - [`"rsa-sha256"`][RSA_SHA256], the most compatible option
#[derive(PartialEq, Debug, Clone, Copy)]
pub struct Algorithm(&'static str);

/// `hs2019`, the only non-deprecated HTTP signatures algorithm.
pub const HS2019: Algorithm = Algorithm("hs2019");
/// The HTTP signatures algorithm everyone uses, `rsa-sha256`.
pub const RSA_SHA256: Algorithm = Algorithm("rsa-sha256");

/// A signature derived from an [`http::Request`].
#[derive(Debug)]
pub struct Signature {
    key_id: String,
    components: Vec<(String, String)>,
    time_range: (DateTime<Utc>, DateTime<Utc>),
    target_str: String,
    signed_str: String,
    signature: String,
    algorithm: Algorithm,
}

impl Signature {
    /// Attempt to extract a signature from a request.
    pub fn derive<T>(req: &Request<T>) -> Result<Signature, String> {
        new::with_ir(req, Options::MODERN, |ir| ir.to_signature())
    }
    /// Obtain the key id for the signature.
    pub fn key_id(&self) -> &str {
        &self.key_id
    }
    /// Get the time the signature was created. This information is extracted from the `(created)`
    /// pseudo-header if it is defined, and the `date` http header otherwise.
    pub fn created(&self) -> DateTime<Utc> {
        self.time_range.0
    }
    /// If specified, get the `(expires)` header, otherwise get the creation time + the configured grace window.
    pub fn expires(&self) -> DateTime<Utc> {
        self.time_range.1
    }
    /// Retrieve the algorithm used for the signature.
    pub fn algorithm(&self) -> Algorithm {
        self.algorithm
    }
    /// Attach `self` to `req` as the `signature` header.
    pub fn commit<T>(self, req: &mut Request<T>) {
        req.headers_mut().insert("signature", self.make_header());
    }
    /// Determine whether `self` was signed by the private counterpart of `key`.
    pub fn was_signed_by(&self, key: &Public) -> Result<bool, String> {
        use rsa::pkcs1v15::{VerifyingKey, Signature};

        let Public(inner) = key.clone();
        let key = VerifyingKey::<Sha256>::new(inner);

        let raw_buf = decode(&self.signature)?;
        let Ok(sig) = Signature::try_from(raw_buf.as_slice()) else {
            do yeet "Failed to construct signature from decoded signature";
        };

        key.verify(self.signed_str.as_bytes(), &sig)
            .map_err(|s| format!("{s:?}"))?;

        Ok(true)
    }
    /// Check whether the given `time` falls within the window for valid signatures.
    pub fn is_expired_at(&self, time: DateTime<Utc>) -> bool {
        !(self.created()..self.expires()).contains(&time)
    }
    /// Turn the signature into an HTTP header value.
    fn make_header(self) -> HeaderValue {
        IR::<&str>::from_signature(&self)
            .to_header()
            .try_into()
            .unwrap()
    }
}

/// `rsa-sha256` is created using an rsa key and a sha256 hash.
fn sign_rsa_sha256(signing_string: &str, key: &Private) -> Result<Vec<u8>, String> {
    use rsa::pkcs1v15::SigningKey;
    let Private(rsa) = key.clone();
    let key = SigningKey::<Sha256>::new(rsa);
    let buf = key.sign(signing_string.as_bytes()).to_vec();
    Ok(buf)
}

/// Maximum time difference between the creation time of the signature and the current time before the
/// signature will be rejected. This is a measure to increase the difficulty of a replay attack.
const EXPIRY_WINDOW: TimeDelta = TimeDelta::minutes(5);

/// Configuration for the behavior of the signing and verification routines.
///
/// This struct is non-exhaustive.
#[derive(Clone, Copy)]
pub struct Options {
    /// Whether to use the `(created)` and `(expires)`. If `false`, the `date` header is used instead.
    ///
    /// Defaults to `true`.
    pub use_created: bool,
    /// Quirk for older mastodon versions, which don't incorporate the query string into the signing
    /// string during verification.
    ///
    /// Defaults to `false`.
    pub strip_query: bool,
    /// For how long the signature is valid.
    ///
    /// For signing, this only has an effect if `use_created` is also set. For verification, this is only
    /// used if the `(expires)` pseudo-header is not present.
    ///
    /// Defaults to 5 minutes.
    pub expires_after: TimeDelta,
    /// Which signature algorithm to use.
    ///
    /// Defaults to [`"hs2019"`][super::HS2019].
    pub algorithm: Algorithm,
}

impl Options {
    /// Use hs2019 with the `(created)` pseudo-header.
    pub const MODERN: Options = Options {
        use_created: true,
        strip_query: false,
        expires_after: EXPIRY_WINDOW,
        algorithm: HS2019,
    };
    /// Use rsa-sha256 with the `date` header.
    pub const LEGACY: Options = Options {
        use_created: false,
        algorithm: RSA_SHA256,
        ..Options::MODERN
    };
}

impl Default for Options {
    fn default() -> Self {
        Options {
            use_created: true,
            strip_query: false,
            expires_after: EXPIRY_WINDOW,
            algorithm: HS2019,
        }
    }
}

mod new {
    use base64::prelude::*;
    use chrono::{DateTime, Utc};
    use http::{Method, Request};
    use rsa::sha2::{Digest, Sha256};

    use super::{sign_rsa_sha256, Options, Signature, SigningKey, EXPIRY_WINDOW, HS2019, RSA_SHA256};

    /// Calculate the SHA256 hash of something.
    pub fn sha256(buf: impl AsRef<[u8]>) -> Vec<u8> {
        <Sha256 as Digest>::digest(buf.as_ref()).to_vec()
    }

    /// Base64-encode something.
    pub fn encode(buf: impl AsRef<[u8]>) -> String {
        BASE64_STANDARD.encode(buf.as_ref())
    }

    /// Base64-decode something.
    pub fn decode(buf: &str) -> Result<Vec<u8>, String> {
        BASE64_STANDARD.decode(buf).map_err(|e| e.to_string())
    }

    pub struct IR<'s, S = &'s str> {
        target: &'s str,
        inputs: Vec<(&'s str, &'s str)>,
        alg: &'s str,
        key: S,
        sig: S,
    }

    /// Allocates a new [`IR`] for doing signature operations with.
    pub fn with_ir<T, U>(
        req: &Request<T>,
        opt: Options,
        f: impl FnOnce(IR<'_>) -> Result<U, String>,
    ) -> Result<U, String> {
        let target = &format_target(&req, opt);
        let map = make_header_map(&req);
        let Some(header) = get(&map, "signature") else {
            do yeet "Missing required `signature` header";
        };
        let (inputs, key, alg, sig) = parse_header(header, &target, &map)?;
        f(IR { target, inputs, key, alg, sig })
    }

    impl<S> IR<'_, S> {
        /// Create an HTTP header from the IR.
        pub fn to_header(&self) -> String
        where
            S: AsRef<str>,
        {
            format_header(&self.inputs, self.key.as_ref(), self.alg, self.sig.as_ref())
        }
        /// Validate and upgrade the IR to a structured signature.
        pub fn to_signature(&self) -> Result<Signature, String>
        where
            S: ToString,
        {
            let times: Result<_, String> = try {
                let date = get(&self.inputs, "date").map(from_rfc822).transpose()?;
                let created = get(&self.inputs, "(created)").map(from_secs).transpose()?;
                let expires = get(&self.inputs, "(expires)").map(from_secs).transpose()?;
                (date, created, expires)
            };
            let (date, created, expires) =
                times.map_err(|e: String| format!("Failed to parse time: {e}"))?;

            let (created, expires) = match (created, expires) {
                (Some(created), None) => (created, created + EXPIRY_WINDOW),
                (Some(created), Some(expires)) => (created, expires),
                (None, _) => {
                    let Some(date) = date else {
                        do yeet "Cannot determine validity window";
                    };
                    (date, date + EXPIRY_WINDOW)
                }
            };

            let algorithm = match self.alg {
                "rsa-sha256" => RSA_SHA256,
                "hs2019" => HS2019,
                a => do yeet format!("Unsupported algorithm {a}"),
            };
            let signed_str = make_signing_string(&self.inputs);
            let components = self
                .inputs
                .iter()
                .map(|(k, v)| (k.to_string(), v.to_string()))
                .collect();
            Ok(Signature {
                key_id: self.key.to_string(),
                signature: self.sig.to_string(),
                time_range: (created, expires),
                target_str: self.target.to_string(),
                components,
                signed_str,
                algorithm,
            })
        }
        /// Create an IR from a signature.
        pub fn from_signature<'s>(sig: &'s Signature) -> IR<'s> {
            IR {
                target: &sig.target_str,
                inputs: sig
                    .components
                    .iter()
                    .map(|(a, b)| (a.as_str(), b.as_str()))
                    .collect(),
                key: &sig.key_id,
                alg: sig.algorithm.0,
                sig: &sig.signature,
            }
        }
        /// Create a signing string.
        pub fn to_signing_str(&self) -> String {
            make_signing_string(&self.inputs)
        }
    }

    impl<'s> IR<'s, ()> {
        /// Create a partial, unsigned IR.
        pub fn partial<'r, T, U>(
            req: &'r Request<T>,
            opt: Options,
            f: impl FnOnce(IR<'_, ()>) -> Result<U, String>,
        ) -> Result<U, String> {
            let map = make_header_map(req);
            let digest = req.method() == Method::POST;

            let expires_after = opt.use_created.then_some(opt.expires_after);
            let created = Utc::now();
            let expires = created + expires_after.unwrap_or(EXPIRY_WINDOW);

            let target = &format_target(&req, opt);
            let created = created.timestamp().to_string();
            let expires = expires.timestamp().to_string();

            // Association list mapping pseudo headers names to concrete values.
            #[rustfmt::skip]
            let pseudo = &[
                ("(request-target)", target.as_str()),
                ("(created)", created.as_str()),
                ("(expired)", expires.as_str()),
            ];

            let inputs = match compute_inputs(&map, pseudo, opt, digest) {
                Err(error) => do yeet format!("computing inputs: {error}"),
                Ok(inputs) => inputs,
            };

            f(IR {
                target,
                inputs,
                alg: opt.algorithm.0,
                key: (),
                sig: (),
            })
        }
        /// Sign a partially constructed IR to make it actually useful.
        pub fn signed(self, key: &'s SigningKey) -> Result<IR<'s, String>, String> {
            let sig_str = self.to_signing_str();
            let signature = match sign_rsa_sha256(&sig_str, &key.inner).map(encode) {
                Err(error) => do yeet format!("RSA error: {error}"),
                Ok(signature) => signature,
            };
            Ok(IR {
                target: self.target,
                inputs: self.inputs,
                alg: self.alg,
                key: key.id.to_string(),
                sig: signature,
            })
        }
    }

    /// With the given options and headers, compute a set of headers and pseudo-headers that (in order) are to be
    /// turned into the signing string.
    fn compute_inputs<'a>(
        headers: &[(&'a str, &'a str)],
        pseudo: &[(&'a str, &'a str)],
        opt: Options,
        use_digest: bool,
    ) -> Result<Vec<(&'a str, &'a str)>, String> {
        // List of input names that we want. Pseudo-headers are ordered before normal headers.
        let needed = ["(request-target)"]
            .into_iter()
            .chain(if opt.use_created {
                vec!["(created)", "(expired)"]
            } else {
                vec!["date"]
            })
            .chain(use_digest.then_some("digest"))
            .chain(["host"]);

        let assoc = |k| {
            get(headers, k)
                .or_else(|| get(&pseudo, k))
                .ok_or_else(|| format!("Missing (pseudo)header `{k}`"))
                .map(|v| (k, v))
        };

        needed.map(assoc).try_collect()
    }

    /// Allocate a `(request-target)` buffer.
    fn format_target<T>(req: &Request<T>, opt: Options) -> String {
        let path = if opt.strip_query {
            req.uri().path()
        } else {
            req.uri()
                .path_and_query()
                .map(|r| r.as_str())
                .unwrap_or_else(|| req.uri().path())
        };
        let method = req.method().as_str().to_ascii_lowercase();
        format!("{method} {path}")
    }

    fn format_header(inputs: &[(&str, &str)], key: &str, alg: &str, sig: &str) -> String {
        // Format all the headers in the order that we used them in the signing string.
        let headers: String = inputs
            .iter()
            .map(|(k, _)| k.as_ref())
            .intersperse(" ")
            .collect();

        // Get the time-based parameters, if they exist.
        let created = get(inputs, "(created)").map(|v| ("created", v));
        let expires = get(inputs, "(expires)").map(|v| ("expires", v));

        // These parameters are always produced.
        #[rustfmt::skip]
        let table = [
            ("keyId",     key),
            ("algorithm", alg),
            ("signature", sig),
            ("headers",   &headers),
        ];

        // Now we need to format the whole shebang.
        table
            .into_iter()
            // `(created)` is part of a newer draft that not everyone implements.
            .chain(created)
            // `(expires)` is optional per the spec
            .chain(expires)
            // Step 1: all the values need to be surrounded by quotes
            .map(|(k, v)| (k, format!(r#""{v}""#)))
            // Step 2. join each pair together
            .map(|(k, v)| format!("{k}={v}"))
            // Step 3. comma separate everything
            .intersperse(",".to_string())
            // Step 4. fold the entire thing into one
            .collect::<String>()
    }

    fn parse_header<'s>(
        header: &'s str,
        target: &'s str,
        extra: &[(&'s str, &'s str)],
    ) -> Result<(Vec<(&'s str, &'s str)>, &'s str, &'s str, &'s str), String> {
        // Parse the top-level table.
        let table: Vec<(&str, &str)> = header
            // Split into entries
            .split(",")
            // Split entries into key-value pairs
            .filter_map(|pair| {
                pair.trim_end_matches(' ') // QUIRK: akkoma does not put a space between entries
                    .split_once('=')
            })
            // Undo quoting of the values
            .filter_map(|(k, v)| v.strip_prefix('"')?.strip_suffix('"').map(|v| (k, v)))
            .collect();

        let Some(headers) = get(&table, "headers") else {
            do yeet "Missing `headers` field";
        };
        let Some(key) = get(&table, "keyId") else {
            do yeet "Missing `keyId` field";
        };
        let Some(algorithm) = get(&table, "algorithm") else {
            do yeet "Missing `algorithm` field";
        };
        let Some(signature) = get(&table, "signature") else {
            do yeet "Missing `signature` field"
        };

        let inputs: Vec<(&str, &str)> = headers
            // Headers and pseudo-headers are separated by spaces in the order in which they appear.
            .split(' ')
            // Map created and expires pseudo-headers to the ones specified in the inputs table.
            .map(|k| match k {
                "(request-target)" => Ok(("(request-target)", target)),
                // If these exist, the table must have them, but other than that they're optional.
                "(created)" => get(&table, "created")
                    .ok_or("`(created)` pseudo-header is listed, but does not exist".to_string())
                    .map(|v| ("(created)", v)),
                "(expires)" => get(&table, "expires")
                    .ok_or("`(expires)` pseudo-header is listed, but does not exist".to_string())
                    .map(|v| ("(expires)", v)),
                // For anything else, we don't have the required information, and we'll need access
                // to the entire request in order to fill in the blanks.
                k => get(&extra, k)
                    .ok_or(format!("header '{k}' is missing"))
                    .map(|v| (k, v)),
            })
            .try_collect()?;

        Ok((inputs, key, algorithm, signature))
    }

    /// Make an association list associating header names to header values.
    ///
    /// Allocates a new vector, but not any strings.
    fn make_header_map<'r, T>(req: &'r Request<T>) -> Vec<(&'r str, &'r str)> {
        req.headers()
            .iter()
            // Acquire string slices of every name-value pair.
            .filter_map(|(k, v)| v.to_str().ok().map(|v| (k.as_str(), v)))
            .collect()
    }

    /// Quick utility function to get stuff from an association list.
    fn get<'x>(map: &[(&str, &'x str)], key: &str) -> Option<&'x str> {
        map.iter()
            .find_map(|(k, v)| k.eq_ignore_ascii_case(key).then_some(*v))
    }

    fn make_signing_string(data: &[(&str, &str)]) -> String {
        data.iter()
            // Each pair is separated by a colon and a space
            .map(|(k, v)| format!("{k}: {v}"))
            // Pairs must be separated by a newline
            .intersperse("\n".to_string())
            .collect()
    }

    fn from_secs(s: &str) -> Result<DateTime<Utc>, String> {
        s.parse::<i64>().map_err(|e| e.to_string()).and_then(|t| {
            let Some(time) = DateTime::from_timestamp(t, 0) else {
                do yeet "Timestamp out of range";
            };
            Ok(time)
        })
    }
    fn from_rfc822(s: &str) -> Result<DateTime<Utc>, String> {
        DateTime::parse_from_rfc2822(s)
            .map_err(|e| e.to_string())
            .map(|time| time.to_utc())
    }
}
