//! If you're an ActivityPub developer looking for information about ActivityPuppy's federation behavior,
//! you should take a look at [`fetch`].

// Working with result types is such a bitch without these.
#![feature(iterator_try_collect, try_blocks, once_cell_try, box_into_inner)]

use std::hint::unreachable_unchecked;

use actor::get_signing_key;
pub use context::Context;
#[cfg(test)]
pub use context::test_context;

use data::{ActivityKind, AuthorOf, Channel, Content, Create, Id, ObjectKind, Profile, PublicKey};

use fetch::object::{Activity, Note, Object};
use store::Transaction;
pub use store::{self, Key, StoreError};
pub use fetch::{self, FetchError};

mod context;
pub mod data;

pub mod post;
mod interact;

use derive_more::{From, Display};
use tracing::{instrument, warn};

/// Retrieve an ActivityPub object from the database.
///
/// Fails with `Error::Missing` if the required properties are not present.
pub fn get_local_ap_object(tx: &Transaction<'_>, key: Key) -> Result<fetch::object::Object> {
    let Some(obj) = tx.get_mixin::<data::Object>(key)? else {
        // We need this data in order to determine the object type. If the passed key does not
        // have this data, it must not be an ActivityPub object.
        return Err(Error::MissingData { node: key, prop: "Object" });
    };
    match obj.kind {
        ObjectKind::Actor => {
            let Some(Profile { account_name, display_name, .. }) = tx.get_mixin(key)? else {
                return Err(Error::MissingData { node: key, prop: "Profile" });
            };
            let Some(Channel { inbox }) = tx.get_mixin(key)? else {
                return Err(Error::MissingData { node: key, prop: "Channel" });
            };
            let Some(PublicKey { key_id, key_pem }) = tx.get_mixin(key)? else {
                return Err(Error::MissingData { node: key, prop: "PublicKey" });
            };
            Ok(fetch::object::Object::Actor(fetch::object::Actor {
                id: obj.id.0.clone().into(),
                inbox: inbox.into(),
                account_name: account_name.0,
                display_name,
                public_key: fetch::object::PublicKey {
                    owner: obj.id.0.into(),
                    id: key_id.into(),
                    inner: key_pem,
                },
            }))
        }
        ObjectKind::Activity(ActivityKind::Create) => {
            let Some(Create { object, actor, .. }) = tx.get_arrow(key)? else {
                panic!("expected a `Create`");
            };
            let Id(actor) = tx.get_alias(actor)?.unwrap();
            Ok(fetch::object::Object::Activity(fetch::object::Activity {
                id: obj.id.0.into(),
                actor: actor.into(),
                object: Box::new(get_local_ap_object(tx, object)?),
                kind: String::from("Create"),
            }))
        }
        ObjectKind::Notelike(kind) => {
            let Some(Content { content, warning, .. }) = tx.get_mixin(key)? else {
                panic!()
            };
            let Some(AuthorOf { author, .. }) = tx.incoming(key).next().transpose()? else {
                panic!()
            };
            let Some(Id(author)) = tx.get_alias(author)? else {
                todo!()
            };
            Ok(fetch::object::Object::Note(Note {
                id: obj.id.0.clone().into(),
                summary: warning,
                content,
                author,
                kind,
            }))
        }
        _ => todo!(),
    }
}

pub mod actor {
    use fetch::{
        object,
        signatures::{Private, SigningKey},
    };
    use store::{Key, StoreError, Transaction};

    use crate::{
        data::{Channel, Id, Object, ObjectKind, PrivateKey, Profile, PublicKey, Username},
        Context, Error, Result,
    };

    /// A reference to an actor.
    #[derive(Clone, Copy, Eq, PartialEq, Debug)]
    pub struct Actor {
        /// The key identifying the actor in the data store.
        pub key: Key,
    }

    impl Actor {
        /// Get a local actor from the store by their username.
        pub fn by_username(tx: &Transaction<'_>, username: impl ToString) -> Result<Option<Actor>> {
            let maybe_key = tx
                .lookup(Username(username.to_string()))
                .map_err(Error::Store)?;
            Ok(maybe_key.map(|key| Actor { key }))
        }
    }

    /// Create a fresh local actor.
    pub fn create_local(cx: &Context, username: impl ToString) -> Result<Actor> {
        let key = Key::gen();
        cx.run(|tx| {
            let username: Username = username.to_string().into();
            // Federation stuff
            mixin_ap_actor(tx, key, &cx.config().ap_domain, true)?;
            mixin_priv_key(tx, key, &cx.config().ap_domain)?;
            // Social properties
            tx.add_alias(key, username.clone())?;
            tx.add_mixin(key, Profile {
                post_count: 0,
                account_name: username,
                display_name: None,
                about_string: None,
                about_fields: Vec::new(),
            })?;
            Ok(Actor { key })
        })
    }

    /// Register an actor from another server.
    pub fn create_remote(tx: &Transaction<'_>, object: object::Actor) -> Result<Actor> {
        let key = Key::gen();
        tx.add_alias(key, Id(object.id.clone()))?;
        tx.add_mixin(key, Channel { inbox: object.inbox })?;
        tx.add_mixin(key, Object {
            kind: ObjectKind::Actor,
            id: Id(object.id),
            local: false,
        })?;
        tx.add_mixin(key, Profile {
            post_count: 0,
            account_name: Username(object.account_name),
            display_name: object.display_name,
            about_string: None,
            about_fields: Vec::new(),
        })?;
        tx.add_mixin(key, PublicKey {
            key_id: object.public_key.id,
            key_pem: object.public_key.inner,
        })?;
        Ok(Actor { key })
    }

    /// Add properties related to local ActivityPub actors to a vertex.
    pub fn mixin_ap_actor(
        tx: &Transaction<'_>,
        vertex: Key,
        domain: &str,
        local: bool,
    ) -> Result<(), StoreError> {
        let id = Id(format!("https://{domain}/o/{vertex}"));
        tx.add_alias(vertex, id.clone())?;
        tx.add_mixin(vertex, Channel { inbox: format!("{id}/inbox") })?;
        tx.add_mixin(vertex, Object {
            kind: ObjectKind::Actor,
            local,
            id,
        })?;
        store::OK
    }

    /// Generate and attach a public/private key pair to the vertex.
    pub fn mixin_priv_key(
        tx: &Transaction<'_>,
        vertex: Key,
        domain: &str,
    ) -> Result<(), StoreError> {
        let key_id = format!("https://{domain}/o/{vertex}#sig-key");
        let (private, public) = Private::gen();
        tx.add_mixin(vertex, PublicKey {
            key_pem: public.encode_pem(),
            key_id,
        })?;
        tx.add_mixin(vertex, PrivateKey { key_pem: private.encode_pem() })?;
        store::OK
    }

    pub fn get_signing_key(tx: &Transaction<'_>, actor: Actor) -> Result<SigningKey, StoreError> {
        let (PrivateKey { key_pem, .. }, PublicKey { key_id, .. }) =
            tx.get_mixin_many(actor.key)?;
        let Id(owner) = tx.get_alias(actor.key)?.unwrap();
        let inner = Private::decode_pem(&key_pem);
        Ok(SigningKey { id: key_id, owner, inner })
    }
}

pub type Result<T, E = Error> = std::result::Result<T, E>;

#[derive(From, Debug, Display)]
pub enum Error {
    /// An error internal to the store.
    #[display(fmt = "store error: {}", self.0)]
    Store(StoreError),
    /// An error generated by the [fetch] subsystem.
    #[display(fmt = "fetch error: {}", self.0)]
    Fetch(FetchError),
    /// Expected `node` to have some property that it doesn't have.
    #[display(fmt = "missing data: {node} is missing {prop}")]
    MissingData {
        /// The node that is missing the data.
        node: Key,
        /// Name of the thing it is missing.
        prop: &'static str,
    },
    #[display(fmt = "invalid data: {}", self.0)]
    Invalid(String),
}

pub mod config {
    #[derive(Clone)]
    pub struct Config {
        pub ap_domain: String,
        pub wf_domain: String,
        pub state_dir: String,
        pub port: u16,
    }
}

/// Interpret an *incoming* activity. Outgoing activities are *never* interpreted through this function,
/// because their changes are already in the database.
// TODO: figure out if that is the behavior we actually want
#[instrument(skip_all, fields(activity.id = activity.id))]
pub fn interpret(cx: &Context, activity: Activity) -> Result<()> {
    // Fetch our actor from the database
    let Some(actor) = cx.store().lookup(Id(activity.actor.clone()))? else {
        panic!(
            "actor {} does not exist in the database (id={})",
            activity.actor, activity.id
        )
    };

    // Fetch our object from the database. The object must already exist in the database.
    let id = activity.object.id();
    let Some(object) = cx.store().lookup(Id(id.to_owned()))? else {
        panic!(
            "object {} does not exist in the database (id={})",
            activity.object.id(),
            activity.id
        )
    };

    let actor = actor::Actor { key: actor };
    let (key, tag) = match activity.kind.as_str() {
        "Bite" => {
            let object = actor::Actor { key: object };
            (actor.do_bite(&cx, object)?.id, ActivityKind::Bite)
        }
        "Create" => {
            // NOTE: due to the ingesting, we already have this information.
            // TODO: change this. for god's sake
            return Ok(());
        }
        "Follow" => {
            let object = actor::Actor { key: object };

            let req = actor.do_follow_request(&cx, object)?;
            (req.id, ActivityKind::Follow)
        }
        tag @ ("Accept" | "Reject") => {
            // Follow requests are multi-arrows in our graph, and they have their own activitypub id.
            let Some(req) = cx.store().get_arrow(object)? else {
                panic!(
                    "follow request does not exist: {object} (id={})",
                    activity.id
                )
            };
            // Dispatch to the actual method based on the tag
            let tag = match tag {
                "Accept" => actor
                    .do_accept_request(&cx, req)
                    .map(|_| ActivityKind::Accept)?,
                "Reject" => actor
                    .do_reject_request(&cx, req)
                    .map(|_| ActivityKind::Reject)?,
                _ => unsafe {
                    // SAFETY: this branch of the outer match only matches if the tag is either "Accept" or "Reject",
                    // so this inner branch is truly unreachable.
                    unreachable_unchecked()
                },
            };
            (Key::gen(), tag)
        }
        k => {
            warn!(activity.id, "unsupported activity type {k}");
            return Err(Error::Invalid(format!("activity type '{k}' not supported")));
        }
    };
    cx.run(|tx| {
        tx.add_alias(key, Id(activity.id.clone()))?;
        tx.add_mixin(key, data::Object {
            id: Id(activity.id.clone()),
            kind: ObjectKind::Activity(tag),
            local: false,
        })?;
        Ok(())
    })
}

/// Make sure all the interesting bits of an activity are here.
#[instrument(skip_all, fields(activity.id = activity.id, key = auth.to_string()))]
pub async fn ingest(cx: &Context, auth: Key, activity: &Activity) -> Result<()> {
    let key = cx.run(|tx| get_signing_key(tx, actor::Actor { key: auth }).map_err(Error::Store))?;
    for id in [activity.actor.as_str(), activity.object.id()] {
        if cx.store().lookup(Id(id.to_owned()))?.is_some() {
            // Skip ingesting if we already know this ID.
            continue;
        }
        let json = cx.resolver().resolve(&key, &id).await?;
        let object = Object::from_json(json).unwrap();
        match object {
            Object::Activity(a) => interpret(&cx, a)?,
            Object::Actor(a) => cx.run(|tx| actor::create_remote(tx, a).map(void))?,
            Object::Note(a) => post::create_post_from_note(cx, a).map(void)?,
            _ => todo!(),
        }
    }
    Ok(())
}

/// Discard the argument.
fn void<T>(_: T) -> () {}
