//! Provides a nice hashmap-esque interface for manipulating entries in the store's backend.

use std::sync::Arc;

use rocksdb::{BoundColumnFamily, IteratorMode};

pub use self::cx::{Context, Query, Write};
use crate::{util::IterExt as _, Result, StoreError};

/// An internal interface to a specific keyspace that exposes basic hashmap-esque operations
/// on that keyspace, generic over whether the source of the data is a [`Transaction`] or a
/// [`Store`].
pub struct Keyspace<'db, C> {
    pub(super) context: &'db C,
    pub(super) cf: Arc<BoundColumnFamily<'db>>,
}

impl<'db, C> Keyspace<'db, C>
where
    C: Query,
{
    /// Fetch a value from the keyspace.
    pub fn get(&self, key: impl AsRef<[u8]>) -> Result<Option<impl AsRef<[u8]> + 'db>> {
        self.context.get_pinned(&self.cf, key)
    }
    /// Test whether a key exists.
    pub fn has(&self, key: impl AsRef<[u8]>) -> Result<bool> {
        self.get(key).map(|r| r.is_some())
    }
    /// Execute a prefix scan.
    pub fn scan(
        &self,
        prefix: impl AsRef<[u8]> + 'db,
    ) -> impl Iterator<Item = Result<(Box<[u8]>, Box<[u8]>)>> + 'db {
        let t = prefix.as_ref().to_vec();
        self.context
            .prefix_iterator(&self.cf, prefix.as_ref())
            // The prefix iterator may "overshoot". This makes it stop when it reaches
            // the end of the range that has the prefix.
            .take_while(move |r| match r {
                Ok((ref k, _)) => k.starts_with(&t),
                _ => true,
            })
            .map_err(StoreError::Internal)
    }
    /// List all pairs in the keyspace.
    pub fn list(&self) -> impl Iterator<Item = Result<(Box<[u8]>, Box<[u8]>)>> + 'db {
        self.context
            .full_iterator(&self.cf, IteratorMode::Start)
            .map_err(StoreError::Internal)
    }
    /// Execute a range scan
    pub fn range<const N: usize>(
        &self,
        lower: [u8; N],
        upper: [u8; N],
    ) -> impl Iterator<Item = Result<(Box<[u8]>, Box<[u8]>)>> + 'db {
        // TODO: use a seek op to make this more efficient
        self.context
            .full_iterator(&self.cf, IteratorMode::Start)
            .skip_while(move |r| match r {
                Ok((ref k, _)) => k.as_ref() < &lower,
                _ => false,
            })
            // The prefix iterator may "overshoot". This makes it stop when it reaches
            // the end of the range that has the prefix.
            .take_while(move |r| match r {
                Ok((ref k, _)) => k.as_ref() < &upper,
                _ => true,
            })
            .map_err(StoreError::Internal)
    }
    /// Join all the keys to their values in this keyspace.
    ///
    /// This may be optimized compared to many random point lookups.
    pub fn join(
        &self,
        keys: impl IntoIterator<Item = impl AsRef<[u8]>>,
    ) -> Vec<Result<Option<Vec<u8>>>> {
        self.context
            .multi_get(keys.into_iter().map(|x| (&self.cf, x)))
    }
}

impl<C> Keyspace<'_, C>
where
    C: Write,
{
    /// Set the given `key` to the `value`, overwriting it if there was already a value there.
    pub fn set(&self, key: impl AsRef<[u8]>, val: impl AsRef<[u8]>) -> Result<()> {
        self.context.put(&self.cf, key, val)
    }
    /// Drop the value if it exists.
    pub fn del(&self, key: impl AsRef<[u8]>) -> Result<()> {
        self.context.delete(&self.cf, key)
    }
}

mod cx {
    //! Contexts for doing reads, writes or both to the database.
    //!
    //! The traits in this module map abstract calls to their methods on the [rocksdb] objects.

    use rocksdb::{
        AsColumnFamilyRef, DBAccess, DBIteratorWithThreadMode, DBPinnableSlice, IteratorMode,
    };

    use super::Keyspace;
    use crate::{util::IterExt as _, Backend, Batch, Result, Store, StoreError, Transaction, OK};

    /// A context for executing database operations.
    pub trait Context {
        /// Open the keyspace identified by `cf`.
        fn open<'cx>(&'cx self, cf: impl AsRef<str>) -> Keyspace<'cx, Self>
        where
            Self: Sized;
    }

    /// A context in which one can read from the data store.
    ///
    /// Specifically, this maps calls to either the transaction or the store's internals without us having
    /// to implement methods for *both* transactions and the store.
    pub trait Query: Context {
        type Backend: DBAccess;
        fn get_pinned<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            key: impl AsRef<[u8]>,
        ) -> Result<Option<DBPinnableSlice<'a>>>;
        fn prefix_iterator<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            prefix: &[u8],
        ) -> DBIteratorWithThreadMode<'a, Self::Backend>;
        fn full_iterator<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            mode: IteratorMode<'a>,
        ) -> DBIteratorWithThreadMode<'a, Self::Backend>;
        /// Optimized multi-point lookup.
        fn multi_get<'a, C: AsColumnFamilyRef + 'a>(
            &'a self,
            keys: impl IntoIterator<Item = (&'a C, impl AsRef<[u8]>)>,
        ) -> Vec<Result<Option<Vec<u8>>>>;
    }

    /// A context in which one can read from and modify the data store.
    pub trait Write: Context {
        fn delete(&self, cf: &impl AsColumnFamilyRef, key: impl AsRef<[u8]>) -> Result<()>;
        fn put(
            &self,
            cf: &impl AsColumnFamilyRef,
            key: impl AsRef<[u8]>,
            val: impl AsRef<[u8]>,
        ) -> Result<()>;
    }

    impl Context for Store {
        fn open<'cx>(&'cx self, cf: impl AsRef<str>) -> Keyspace<'cx, Self> {
            let name = cf.as_ref();
            let Some(cf) = self.inner.cf_handle(name) else {
                panic!("unregistered keyspace {name}! is it in the schema?")
            };
            Keyspace { context: &self, cf }
        }
    }

    impl Query for Store {
        type Backend = Backend;

        fn get_pinned<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            key: impl AsRef<[u8]>,
        ) -> Result<Option<DBPinnableSlice<'a>>> {
            self.inner
                .get_pinned_cf(cf, key)
                .map_err(StoreError::Internal)
        }

        fn prefix_iterator<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            prefix: &[u8],
        ) -> DBIteratorWithThreadMode<'a, Backend> {
            self.inner.prefix_iterator_cf(cf, prefix)
        }

        fn full_iterator<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            mode: IteratorMode<'a>,
        ) -> DBIteratorWithThreadMode<'a, Self::Backend> {
            self.inner.full_iterator_cf(cf, mode)
        }

        fn multi_get<'a, C: AsColumnFamilyRef + 'a>(
            &'a self,
            keys: impl IntoIterator<Item = (&'a C, impl AsRef<[u8]>)>,
        ) -> Vec<Result<Option<Vec<u8>>>> {
            self.inner
                .multi_get_cf(keys)
                .into_iter()
                .map_err(StoreError::Internal)
                .collect()
        }
    }

    impl Context for Transaction<'_> {
        fn open<'cx>(&'cx self, cf: impl AsRef<str>) -> Keyspace<'cx, Self> {
            let name = cf.as_ref();
            let Some(cf) = self.store.inner.cf_handle(name) else {
                panic!("unregistered keyspace {name}! is it in the schema?")
            };
            Keyspace { context: &self, cf }
        }
    }

    impl<'db> Query for Transaction<'db> {
        type Backend = rocksdb::Transaction<'db, Backend>;

        fn get_pinned<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            key: impl AsRef<[u8]>,
        ) -> Result<Option<DBPinnableSlice<'a>>> {
            self.inner
                .get_pinned_cf(cf, key)
                .map_err(StoreError::Internal)
        }

        fn prefix_iterator<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            prefix: &[u8],
        ) -> DBIteratorWithThreadMode<'a, Self::Backend> {
            self.inner.prefix_iterator_cf(cf, prefix)
        }

        fn full_iterator<'a>(
            &'a self,
            cf: &impl AsColumnFamilyRef,
            mode: IteratorMode<'a>,
        ) -> DBIteratorWithThreadMode<'a, Self::Backend> {
            self.inner.full_iterator_cf(cf, mode)
        }

        fn multi_get<'a, C: AsColumnFamilyRef + 'a>(
            &'a self,
            keys: impl IntoIterator<Item = (&'a C, impl AsRef<[u8]>)>,
        ) -> Vec<Result<Option<Vec<u8>>>> {
            self.inner
                .multi_get_cf(keys)
                .into_iter()
                .map_err(StoreError::Internal)
                .collect()
        }
    }

    impl Write for Transaction<'_> {
        fn delete(&self, cf: &impl AsColumnFamilyRef, key: impl AsRef<[u8]>) -> Result<()> {
            self.inner.delete_cf(cf, key).map_err(StoreError::Internal)
        }

        fn put(
            &self,
            cf: &impl AsColumnFamilyRef,
            key: impl AsRef<[u8]>,
            val: impl AsRef<[u8]>,
        ) -> Result<()> {
            self.inner
                .put_cf(cf, key, val)
                .map_err(StoreError::Internal)
        }
    }

    impl Context for Batch {
        fn open<'cx>(&'cx self, cf: impl AsRef<str>) -> Keyspace<'cx, Self>
        where
            Self: Sized,
        {
            let name = cf.as_ref();
            let Some(cf) = self.store.inner.cf_handle(name) else {
                panic!("unregistered keyspace {name}! is it in the schema?")
            };
            Keyspace { context: &self, cf }
        }
    }

    impl Write for Batch {
        fn delete(&self, cf: &impl AsColumnFamilyRef, key: impl AsRef<[u8]>) -> Result<()> {
            self.inner.borrow_mut().delete_cf(cf, key);
            OK
        }

        fn put(
            &self,
            cf: &impl AsColumnFamilyRef,
            key: impl AsRef<[u8]>,
            val: impl AsRef<[u8]>,
        ) -> Result<()> {
            self.inner.borrow_mut().put_cf(cf, key, val);
            OK
        }
    }
}
