use std::{
    fmt::{Debug, Display},
    str::FromStr,
};

use chrono::{DateTime, Utc};
use ulid::Ulid;

use crate::StoreError;

/// A unique identifier for vertices in the database.
#[derive(Clone, Copy, Eq, PartialEq, Ord, PartialOrd, Hash)]
pub struct Key(pub(crate) [u8; 16]);

impl Key {
    /// Generate a new node identifier.
    pub fn gen() -> Key {
        Key(ulid::Ulid::new().to_bytes())
    }
    /// Get the time at which this key was generated.
    pub fn timestamp(self) -> DateTime<Utc> {
        let ms = self.to_ulid().timestamp_ms();
        DateTime::from_timestamp_millis(ms as i64).unwrap()
    }
    /// Join two keys together.
    pub(crate) fn fuse(self, other: Key) -> [u8; 32] {
        let mut buf = [0; 32];
        buf[..16].copy_from_slice(&self.0);
        buf[16..].copy_from_slice(&other.0);
        buf
    }
    pub(crate) fn from_slice(buf: &[u8]) -> Key {
        let mut key = [0; 16];
        key.copy_from_slice(&buf);
        Key(key)
    }
    pub(crate) fn split(buf: &[u8]) -> (Key, Key) {
        let tail = Key::from_slice(&buf[..16]);
        let head = Key::from_slice(&buf[16..]);
        (tail, head)
    }
    // TODO: This doesn't belong here lmao
    pub(crate) fn range(ts: DateTime<Utc>) -> ([u8; 16], [u8; 16]) {
        let min = Ulid::from_parts(ts.timestamp_millis() as u64, u128::MIN).to_bytes();
        let max = Ulid::from_parts(ts.timestamp_millis() as u64, u128::MAX).to_bytes();
        (min, max)
    }
    fn to_ulid(self) -> Ulid {
        Ulid::from_bytes(self.0)
    }
}

impl AsRef<[u8]> for Key {
    fn as_ref(&self) -> &[u8] {
        &self.0
    }
}

impl FromStr for Key {
    type Err = crate::StoreError;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        s.parse::<Ulid>()
            .map(|x| Key(x.to_bytes()))
            .map_err(|err| StoreError::BadKey(err))
    }
}

impl Display for Key {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        Display::fmt(&self.to_ulid(), f)
    }
}

impl Debug for Key {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "Key({})", Ulid::from_bytes(self.0))
    }
}
