#![feature(iterator_try_collect, associated_type_defaults)]
//! Data persistence for the ActivityPuppy social media server built on top of [rocksdb].
//!
//! # Overview
//!
//! The design of the data store's abstractions is heavily inspired by graph theory. The idea is to encourage
//! composition and loose coupling by segmenting all data associated with a node into [mixins][Mixin], and
//! modeling relations and predicates between nodes as [arrows][Arrow]. In additions, the key identifying a
//! node can be [aliased][Alias] by a string newtype, which must be unique within the namespace of that alias.
//!
//! The API is optimized for reducing boilerplate and legibility at the call site.
//!
//! There are three interfaces to the store: the read-only [`Store`], the write-only [`Batch`] and the [`Transaction`],
//! which allows both reads and writes.

use std::{cell::RefCell, future::Future, path::Path, sync::Arc};

use derive_more::{From, Display};
use rocksdb::{Options, TransactionDBOptions, WriteBatchWithTransaction};
use types::Schema;

mod alias;
mod internal;
mod key;
mod mixin;

/// The name of the puppy data store inside the state directory.
const STORE_NAME: &str = "main-store";

/// Open a [`Store`]. Creates one if it doesn't exist yet at the state directory path.
pub fn open(state_dir: impl AsRef<Path>, schema: Schema) -> Result<Store> {
    Store::open(state_dir, schema)
}

pub use {alias::Alias, arrow::Arrow, key::Key, mixin::Mixin};

pub mod arrow;
pub mod types;
pub mod util;

/// The main interface to the data persistence engine.
///
/// This type provides reading capabilities, but does not expose APIs for manipulating data in the store. For
/// that, you must [run][Store::run] a [`Transaction`] or [apply][Store::apply] a [`Batch`].
#[derive(Clone)]
pub struct Store {
    // TODO: maybe switch to `OptimisticTransactionDB` because it has `batched_multi_get_cf`, which may be useful
    // if we end up doing lots of point lookups. alternatively, maybe we don't need *transactions* altogether, and
    // we can get away with write batches and snapshots. the main problem with transactions is that it doesn't let
    // us do range deletes, which affects the efficiency of multiarrow deletion.
    //
    // a switch to write batches is feasible if we end up not doing reads and writes in the same transaction.
    inner: Arc<Backend>,
}

/// Hosts APIs for manipulating the data store.
///
/// You can access these APIs from the body of the closure passed to [`Store::run`].
pub struct Transaction<'db> {
    inner: rocksdb::Transaction<'db, Backend>,
    store: &'db Store,
}

/// A set of writes that are to be executed atomically.
pub struct Batch {
    inner: RefCell<WriteBatchWithTransaction<true>>,
    store: Store,
}

impl Store {
    /// Run a [transaction][Transaction].
    ///
    /// In a transaction, either all writes succeed, or the transaction is aborted and the changes are not
    /// recorded. Changes made inside a transaction can be read from within that transaction before they are
    /// committed.
    ///
    /// If the closure passed to `run` returns an error, the transaction is rolled back, and otherwise the
    /// changes are committed.
    pub fn run<T, E>(&self, f: impl FnOnce(&Transaction<'_>) -> Result<T, E>) -> Result<T, E>
    where
        E: From<StoreError>,
    {
        let tx = Transaction {
            inner: self.inner.transaction(),
            store: &self,
        };
        let r = f(&tx);
        if let Err(e) = if r.is_err() {
            tx.inner.rollback()
        } else {
            tx.inner.commit()
        } {
            return Err(E::from(StoreError::Internal(e)));
        }
        r
    }
    /// Apply a batch of changes atomically.
    pub fn apply(&self, batch: Batch) -> Result<()> {
        self.inner.write(batch.inner.into_inner())?;
        OK
    }
    /// Construct a [`Batch`].
    pub fn batch(&self) -> Batch {
        Batch {
            inner: RefCell::new(WriteBatchWithTransaction::default()),
            store: self.clone(),
        }
    }
    /// Open the data store in `state_dir`, and create one if it doesn't exist yet.
    pub fn open(state_dir: impl AsRef<Path>, schema: Schema) -> Result<Store> {
        let mut db_opts = Options::default();
        db_opts.create_if_missing(true);
        db_opts.create_missing_column_families(true);
        let tx_opts = TransactionDBOptions::default();
        let inner = Arc::new(Backend::open_cf(
            &db_opts,
            &tx_opts,
            state_dir.as_ref().join(STORE_NAME),
            schema.0,
        )?);
        Ok(Store { inner })
    }
    /// Delete the main data store in `state_dir` if it exists.
    pub fn nuke(state_dir: impl AsRef<Path>) -> Result<()> {
        Backend::destroy(&Options::default(), state_dir.as_ref().join(STORE_NAME))
            .map_err(StoreError::Internal)
    }
    /// Open a store that lives until `f` returns, for testing.
    pub fn test<T>(schema: Schema, f: impl FnOnce(Store) -> T) -> T {
        let tmp_dir = tempfile::tempdir().expect("couldn't create tempdir");
        f(Store::open(tmp_dir, schema).expect("failed to open temporary data store in {tmp_dir}"))
    }
}

/// A shorthand for committing a [`Transaction`] (because I think `Ok(())` is ugly).
pub const OK: Result<()> = Ok(());

/// Results from this component.
pub type Result<T, E = StoreError> = std::result::Result<T, E>;

/// Errors from the data store.
#[derive(From, Display, Debug)]
pub enum StoreError {
    /// The requested value was expected to exist in a particular keyspace, but does not actually
    /// exist there. This can occur on updates for example.
    Missing,
    /// One or more operand keys are not registered in the database.
    Undefined {
        key: Key,
    },
    /// Returned if there is a conflict; for example, if the uniqueness property of an alias would
    /// be violated by inserting one.
    Conflict,
    /// A node key couldn't be decoded.
    BadKey(ulid::DecodeError),
    /// Signals a failure related to the data store's backend.
    Internal(rocksdb::Error),
    Encoding(bincode::error::EncodeError),
    Decoding(bincode::error::DecodeError),
}

type Backend = rocksdb::TransactionDB<rocksdb::MultiThreaded>;
