//! Utilities for making the code clearer.

/// Helpers for working with iterators of results of key-value pairs.
pub trait IterExt: Iterator + Sized {
    /// Map over the values inside the iterator's results.
    fn map_ok<'a, I, O, E>(
        self,
        mut f: impl FnMut(I) -> O + 'a,
    ) -> impl Iterator<Item = Result<O, E>> + 'a
    where
        Self: Iterator<Item = Result<I, E>> + 'a,
    {
        self.map(move |r| r.map(|x| f(x)))
    }
    /// Map over the errors contained in the iterator.
    fn map_err<'a, I, O, E>(
        self,
        mut f: impl FnMut(E) -> O + 'a,
    ) -> impl Iterator<Item = Result<I, O>> + 'a
    where
        Self: Iterator<Item = Result<I, E>> + 'a,
    {
        self.map(move |r| r.map_err(|x| f(x)))
    }
    /// Discard the second element of the tuple.
    fn keys<'a, K, V, E>(self) -> impl Iterator<Item = Result<K, E>> + 'a
    where
        Self: Iterator<Item = Result<(K, V), E>> + 'a,
    {
        self.map_ok(|(k, _)| k)
    }
    /// Like and_then.
    fn bind_results<'a, I, O, E>(
        self,
        mut f: impl FnMut(I) -> Result<O, E> + 'a,
    ) -> impl Iterator<Item = Result<O, E>> + 'a
    where
        Self: Iterator<Item = Result<I, E>> + 'a,
    {
        self.map(move |r| r.and_then(|x| f(x)))
    }
    /// Get the next element, or return the given error.
    fn next_or<'a, T, E>(&mut self, e: E) -> Result<T, E>
    where
        Self: Iterator<Item = Result<T, E>> + 'a,
    {
        self.next().ok_or(e)?
    }
    /// `filter_map` meets `and_then`.
    fn filter_bind_results<'a, I, O, E>(
        self,
        mut f: impl FnMut(I) -> Result<Option<O>, E> + 'a,
    ) -> impl Iterator<Item = Result<O, E>> + 'a
    where
        Self: Iterator<Item = Result<I, E>> + 'a,
    {
        self.filter_map(move |r| r.and_then(|x| f(x)).transpose())
    }
    /// Like [`Iterator::find`].
    fn find_ok<'a, I, E>(mut self, mut f: impl FnMut(&I) -> bool) -> Result<Option<I>, E>
    where
        Self: Iterator<Item = Result<I, E>> + 'a,
    {
        self.find(move |r| r.as_ref().is_ok_and(|x| f(x)))
            .transpose()
    }
}

impl<I> IterExt for I where I: Iterator {}

/// Curried function for creating a tuple, where the first argument is the first tuple element, and the next
/// argument is the second tuple element.
pub fn key<K, V>(key: K) -> impl FnOnce(V) -> (K, V) {
    move |val| (key, val)
}

/// Curried function for creating a tuple, where the first argument is the first tuple element, and the next
/// argument is the second tuple element.
pub fn val<K, V>(val: V) -> impl FnOnce(K) -> (K, V) {
    move |key| (key, val)
}
